package limax.auany;

import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;
import java.net.SocketAddress;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.sun.net.httpserver.HttpHandler;

import limax.auany.plats.Experience;
import limax.auany.switcherauany.SessionAuthByToken;
import limax.defines.ErrorCodes;
import limax.defines.ErrorSource;
import limax.defines.SessionFlags;
import limax.endpoint.AuanyService.ResultExperienceUpgrade;
import limax.switcherauany.AuanyAuthArg;
import limax.switcherauany.AuanyAuthRes;
import limax.util.Trace;
import limax.zdb.Procedure;

final class PlatManager {
	private final static Map<String, PlatProcess> plats = new ConcurrentHashMap<>();

	static void initialize(Element self, Map<String, HttpHandler> httphandlers) throws Exception {
		NodeList list = self.getElementsByTagName("plat");
		int count = list.getLength();
		for (int i = 0; i < count; i++)
			parsePlatElement((Element) list.item(i), httphandlers);
		LoginInfoCache.initialize(self);
	}

	static void unInitialize() {
		LoginInfoCache.unInitialize();
	}

	private static void parsePlatElement(Element e, Map<String, HttpHandler> httphandlers)
			throws ClassNotFoundException, InstantiationException, IllegalAccessException {
		String classname = e.getAttribute("className");
		String platname = e.getAttribute("name").toUpperCase();
		if (null == classname || classname.isEmpty())
			return;
		if (Trace.isDebugEnabled())
			Trace.debug("Config.load plat classname = " + classname + " platname = " + platname);
		Class<?> cls = Class.forName(classname);
		PlatProcess process = (PlatProcess) cls.newInstance();
		process.init(e, httphandlers);
		if (null != plats.put(platname.toLowerCase(), process))
			throw new RuntimeException("duplicate play process type " + platname);
	}

	private PlatManager() {
	}

	static byte[] onHttpServerCallBack(String path, String query) {
		return null;
	}

	private static boolean checkClientAddress(SocketAddress peeraddress, SocketAddress reportaddress, int type) {
		return true;
	}

	private static void response(SessionAuthByToken rpc) {
		try {
			rpc.response();
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error(rpc, e);
		}
	}

	static void process(SessionAuthByToken rpc) {
		AuanyAuthArg arg = rpc.getArgument();
		AuanyAuthRes res = rpc.getResult();
		PlatProcess pp = plats.get(arg.platflag = arg.platflag.toLowerCase());
		if (null == pp) {
			res.errorSource = ErrorSource.LIMAX;
			res.errorCode = ErrorCodes.AUANY_UNKNOWN_PLAT;
			response(rpc);
			return;
		}
		if (!pp.isOTP()) {
			AuanyAuthRes resCached = LoginInfoCache.tryLogin(arg);
			if (resCached != null) {
				rpc.setResult(resCached);
				response(rpc);
				return;
			}
		}
		SocketAddress peeraddress, reportaddress;
		try (final ObjectInputStream ois = new ObjectInputStream(
				new ByteArrayInputStream(arg.clientaddress.array(), 0, arg.clientaddress.size()))) {
			peeraddress = (SocketAddress) ois.readObject();
			reportaddress = (SocketAddress) ois.readObject();
		} catch (Exception e) {
			peeraddress = reportaddress = new SocketAddress() {
				private static final long serialVersionUID = -210184168558204443L;
			};
		}
		if (!checkClientAddress(peeraddress, reportaddress, arg.sockettype)) {
			res.errorSource = ErrorSource.LIMAX;
			res.errorCode = ErrorCodes.AUANY_CHECK_LOGIN_IP_FAILED;
			response(rpc);
			return;
		}
		pp.checkUserToken(arg.username, arg.token, (userid, errorSource, errorCode) -> {
			if (ErrorCodes.SUCCEED != errorCode || ErrorSource.LIMAX != errorSource) {
				res.errorSource = errorSource;
				res.errorCode = errorCode;
				response(rpc);
				return;
			}
			Procedure.execute(new GetSessionProcedure(arg.platflag, pp instanceof Experience, arg.username, userid,
					arg.pvids.keySet()), (p, r) -> {
				if (r.isSuccess()) {
					res.sessionid = p.getSessionId();
					res.flags = p.getFlags();
					res.permissions.addAll(p.getPermission());
					res.errorSource = ErrorSource.LIMAX;
					res.errorCode = ErrorCodes.SUCCEED;
					LoginInfoCache.add(rpc);
				} else {
					res.errorSource = ErrorSource.LIMAX;
					res.errorCode = ErrorCodes.AUANY_CALL_PROCEDURE_FAILED;
				}
				response(rpc);
			});
		});
	}

	static void onExperienceUpgrade(ResultExperienceUpgrade onresult, long sessionid, String username, String token,
			String platflag) {
		final PlatProcess pp = plats.get(platflag.toLowerCase());
		if (null == pp) {
			onresult.accept(ErrorSource.LIMAX, ErrorCodes.AUANY_UNKNOWN_PLAT);
			return;
		}
		if (pp instanceof Experience) {
			onresult.accept(ErrorSource.LIMAX, ErrorCodes.AUANY_SERVICE_EXPERIENCE_UPGRADE_DST_ACCOUNT_IS_EXPE);
			return;
		}
		try {
			pp.checkUserToken(username, token, (userid, errorSource, errorCode) -> {
				if (ErrorCodes.SUCCEED != errorCode || ErrorSource.LIMAX != errorSource) {
					onresult.accept(errorSource, errorCode);
					return;
				}
				Procedure.execute(() -> {
					cbean.SessionKey key = table.Sessions.select(sessionid);
					if (null == key) {
						onresult.accept(errorSource,
								ErrorCodes.AUANY_SERVICE_EXPERIENCE_UPGRADE_SRC_ACCOUNT_NOT_EXISTING);
						return false;
					}
					xbean.SessionInfo info = table.Sessioninfos.select(key);
					if (SessionFlags.FLAG_EXPERIENCE_ACCOUNT != (info.getFlags()
							& SessionFlags.FLAG_EXPERIENCE_ACCOUNT)) {
						onresult.accept(errorSource, ErrorCodes.AUANY_SERVICE_EXPERIENCE_UPGRADE_SRC_ACCOUNT_NOT_EXPE);
						return false;
					}
					final cbean.SessionKey newkey = new cbean.SessionKey(platflag, userid);
					if (null != table.Sessioninfos.select(newkey)) {
						onresult.accept(errorSource, ErrorCodes.AUANY_SERVICE_EXPERIENCE_UPGRADE_DST_ACCOUNT_EXISTING);
						return false;
					}
					table.Sessions.delete(sessionid);
					table.Sessions.insert(sessionid, newkey);
					xbean.SessionInfo newinfo = table.Sessioninfos.insert(newkey);
					newinfo.setFlags(info.getFlags() & ~SessionFlags.FLAG_EXPERIENCE_ACCOUNT);
					newinfo.setSessionid(sessionid);
					newinfo.setUsername(username);
					table.Sessioninfos.delete(key);
					onresult.accept(ErrorSource.LIMAX, ErrorCodes.SUCCEED);
					return true;
				});
			});
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error("PlatManager onExperienceUpgrade", e);
		}

	}
}
