package limax.provider;

import java.util.Collection;
import java.util.Collections;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;
import java.util.stream.Stream;

import limax.codec.CodecException;
import limax.net.ClientListener;
import limax.net.ClientManager;
import limax.net.ClientManagerConfig;
import limax.net.Config;
import limax.net.Engine;
import limax.net.Manager;
import limax.net.ManagerConfig;
import limax.net.Protocol;
import limax.net.ServerListener;
import limax.net.SizePolicyException;
import limax.net.Transport;
import limax.provider.switcherprovider.Kick;
import limax.util.Trace;

class ProviderSwitcherExchanger implements ServerListener, ClientListener {
	private final ProviderManagerImpl provider;
	private final Set<Transport> transports = Collections.newSetFromMap(new ConcurrentHashMap<Transport, Boolean>());

	class SessionObject {
		private Set<Long> sessionids = Collections.newSetFromMap(new ConcurrentHashMap<Long, Boolean>());

		void add(long sid) {
			sessionids.add(sid);
		}

		void remove(long sid) {
			sessionids.remove(sid);
		}

		boolean contains(long sid) {
			return sessionids.contains(sid);
		}

		Stream<Long> stream() {
			return sessionids.stream();
		}

		void forEach(Consumer<? super Long> action) {
			sessionids.forEach(action);
		}

		ProviderManagerImpl getProviderManager() {
			return provider;
		}
	}

	ProviderSwitcherExchanger(ProviderManagerImpl provider) {
		this.provider = provider;
	}

	Manager add(ManagerConfig config) throws Exception {
		try {
			return Engine.add(config, this, provider);
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error("ProviderSwitcherListener.add " + config, e);
			throw e;
		}
	}

	@Override
	public void onManagerInitialized(Manager manager, Config config) {
	}

	@Override
	public void onManagerUninitialized(Manager manager) {
	}

	@Override
	public void onTransportAdded(Transport transport) throws Exception {
		try {
			provider.tryBind(transport);
		} catch (Exception e) {
			Manager manager = transport.getManager();
			if (Trace.isErrorEnabled())
				Trace.error(manager.getConfig() + " trybind exception", e);
			if (manager instanceof ClientManager)
				add((ClientManagerConfig) manager.getConfig());
			throw e;
		}
		transport.setSessionObject(new SessionObject());
		transports.add(transport);
		if (Trace.isInfoEnabled())
			Trace.info(provider.getListener() + " onTransportAdded " + transport);
	}

	@Override
	public void onTransportRemoved(Transport transport) throws Exception {
		transports.remove(transport);
		if (provider.isStateless())
			((SessionObject) transport.getSessionObject())
					.forEach(sid -> provider.scheduleSessionTask(sid, () -> provider.statelessRemoveTransport(sid)));
		else
			((SessionObject) transport.getSessionObject())
					.forEach(sid -> provider.scheduleSessionTask(sid, () -> provider.removeTransport(sid)));
		if (Trace.isInfoEnabled())
			Trace.info(provider.getListener() + " onTransportRemoved " + transport, transport.getCloseReason());
	}

	@Override
	public void onAbort(Transport transport) throws Exception {
		if (Trace.isWarnEnabled())
			Trace.warn(provider.getListener() + " onAbort " + transport.getManager().getConfig() + " "
					+ transport.getCloseReason());
	}

	void addSession(Transport transport, long sid) {
		((SessionObject) transport.getSessionObject()).add(sid);
	}

	void removeSession(Transport transport, long sid) {
		((SessionObject) transport.getSessionObject()).remove(sid);
	}

	private void send(Protocol protocol, Transport transport)
			throws InstantiationException, ClassCastException, SizePolicyException, CodecException {
		try {
			protocol.send(transport);
		} catch (InstantiationException | ClassCastException | SizePolicyException | CodecException e) {
			if (Trace.isWarnEnabled())
				Trace.warn("send " + protocol + " on " + transport, e);
			throw e;
		}
	}

	void send(Protocol protocol) throws CodecException {
		CodecException exception = null;
		for (Transport transport : transports)
			try {
				send(protocol, transport);
			} catch (Exception e) {
				if (exception == null)
					exception = new CodecException();
				exception.addSuppressed(e);
			}
		if (exception != null)
			throw exception;
	}

	void send(Protocol protocol, long sid)
			throws InstantiationException, ClassCastException, SizePolicyException, CodecException {
		send(protocol, transports.stream().filter(t -> ((SessionObject) t.getSessionObject()).contains(sid)).findAny()
				.orElseThrow(() -> new CodecException("send " + protocol + " on " + sid + " miss transport")));
	}

	void send(Protocol protocol, Collection<Long> ids) throws CodecException {
		CodecException exception = null;
		for (Transport transport : transports) {
			SessionObject so = (SessionObject) transport.getSessionObject();
			if (ids.stream().filter(id -> so.contains(id)).findAny().isPresent())
				try {
					send(protocol, transport);
				} catch (InstantiationException | ClassCastException | SizePolicyException | CodecException e) {
					if (exception == null)
						exception = new CodecException();
					exception.addSuppressed(e);
				}
		}
		if (exception != null)
			throw exception;
	}

	void kick(long sid, int errorcode) {
		try {
			send(new Kick(sid, errorcode), sid);
		} catch (InstantiationException | ClassCastException | SizePolicyException | CodecException e) {
		}
	}
}
