package limax.provider;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import limax.codec.StringStream;
import limax.provider.providerendpoint.SyncViewToClients;
import limax.providerendpoint.ViewMemberData;
import limax.providerendpoint.ViewVariableData;
import limax.util.Resource;
import limax.util.Trace;

public abstract class TemporaryView extends AutoView {
	public interface MemberShip {
		public enum AbortReason {
			SESSIONMISS, SESSIONCLOSED, DUPLICATEID, MISSID, ADDCANCELLED, VIEWCLOSED, BADREASON
		}

		void add(long sessionid);

		void remove(long sessionid, byte reason);
	}

	static class MemberData {
		final long sessionid;
		final short classindex;
		final DataItem data;

		MemberData(long sessionid, short classindex, DataItem data) {
			this.sessionid = sessionid;
			this.classindex = classindex;
			this.data = data;
		}
	}

	private final static byte DetachReasonClose = -1;
	private final static AtomicInteger indexgenerator = new AtomicInteger();
	private final int instanceindex;
	private final Map<Short, Map<Byte, Byte>> subscribes;
	private final Map<Long, Map<Integer, DataItem>> members = new HashMap<>();
	private final Collection<MemberData> mbcache = new ArrayList<>();
	private final Set<Long> inprogress = new HashSet<>();
	private final MemberShip membership = new MemberShip() {
		private void abortAttach(long sessionid, MemberShip.AbortReason reason) {
			inprogress.remove(sessionid);
			onAttachAbort(sessionid, reason);
		}

		private void abortDetach(long sessionid, MemberShip.AbortReason reason) {
			onDetachAbort(sessionid, reason);
		}

		@Override
		public void add(long sessionid) {
			hashSchedule(() -> add0(sessionid));
		}

		private void add0(long sessionid) {
			inprogress.add(sessionid);
			hashSchedule(sessionid, () -> add1(sessionid));
		}

		private void add1(long sessionid) {
			ViewSession vs = getViewContext().getViewSession(sessionid);
			if (null == vs)
				hashSchedule(() -> abortAttach(sessionid, AbortReason.SESSIONMISS));
			else if (vs.isClosed())
				hashSchedule(() -> abortAttach(sessionid, AbortReason.SESSIONCLOSED));
			else {
				vs.add(TemporaryView.this);
				Map<Short, Collection<DataItem>> initData = new HashMap<>();
				subscribes.forEach(
						(k, v) -> initData.put(k, vs.findSessionView(k).subscribe(TemporaryView.this, v.keySet())));
				hashSchedule(() -> add2(sessionid, vs, initData));
			}
		}

		private void add2(long sessionid, ViewSession vs, Map<Short, Collection<DataItem>> initData) {
			if (isClosed())
				abortAttach(sessionid, AbortReason.VIEWCLOSED);
			else if (members.containsKey(sessionid))
				abortAttach(sessionid, AbortReason.DUPLICATEID);
			else if (!inprogress.contains(sessionid)) {
				hashSchedule(sessionid, () -> vs.remove(TemporaryView.this));
				abortAttach(sessionid, AbortReason.ADDCANCELLED);
			} else {
				if (!members.isEmpty()) {
					flush();
					sendAttachNotify(sessionid, initData);
				}
				Map<Integer, DataItem> memberData = new HashMap<>();
				initData.forEach((k, v) -> v.forEach(v0 -> memberData.put(((int) k << 16) | v0.data.index, v0)));
				members.put(sessionid, memberData);
				sendInitViewToClient(sessionid);
				try {
					onAttached(sessionid);
				} catch (Throwable t) {
					close(sessionid);
				}
				inprogress.remove(sessionid);
			}
		}

		@Override
		public void remove(long sessionid, byte reason) {
			hashSchedule(sessionid, () -> remove0(sessionid, reason));
		}

		private void remove0(long sessionid, byte reason) {
			ViewSession vs = getViewContext().getViewSession(sessionid);
			if (null == vs)
				hashSchedule(() -> abortDetach(sessionid, AbortReason.SESSIONMISS));
			else
				hashSchedule(() -> remove1(sessionid, reason, vs, vs.isClosed()));
		}

		private void remove1(long sessionid, byte reason, ViewSession vs, boolean closed) {
			if (reason < 0)
				abortDetach(sessionid, AbortReason.BADREASON);
			else if (isClosed())
				abortDetach(sessionid, AbortReason.VIEWCLOSED);
			else if (!members.containsKey(sessionid))
				abortDetach(sessionid, AbortReason.MISSID);
			else {
				flush();
				members.remove(sessionid);
				if (!closed)
					sendCloseToClients(Arrays.asList(sessionid));
				if (!members.isEmpty())
					sendDetachNotify(members.keySet(), sessionid, reason);
				hashSchedule(sessionid, () -> remove2(sessionid, vs));
				onDetached(sessionid, reason);
			}
		}

		private void remove2(long sessionid, ViewSession vs) {
			vs.remove(TemporaryView.this);
			unsubscribe(sessionid);
		}
	};

	protected TemporaryView(ViewContext context, ViewStub stub, Resource resource, String[] touch,
			Map<Short, Map<Byte, Byte>> subscribes) {
		super(context, stub, resource, touch);
		this.instanceindex = indexgenerator.incrementAndGet();
		this.subscribes = subscribes;
	}

	@Override
	public void hashSchedule(Runnable task) {
		getViewContext().hashSchedule(this, task);
	}

	private void hashSchedule(long sessionid, Runnable task) {
		getViewContext().hashSchedule(sessionid, task);
	}

	@Override
	Collection<Long> getSessionIds() {
		return members.keySet();
	}

	@Override
	byte getSyncDataType() {
		return SyncViewToClients.DT_TEMPORARY_DATA;
	}

	@Override
	public final int getInstanceIndex() {
		return instanceindex;
	}

	@Override
	public String toString() {
		return "[ class = " + getClass().getName() + " ProviderId = " + getProviderId() + " classindex = "
				+ getClassIndex() + " instanceName = " + getInstanceIndex() + "]";
	}

	public final MemberShip getMemberShip() {
		return membership;
	}

	protected abstract void onAttachAbort(long sessionid, MemberShip.AbortReason reason);

	protected abstract void onDetachAbort(long sessionid, MemberShip.AbortReason reason);

	protected abstract void onAttached(long sessionid);

	protected abstract void onDetached(long sessionid, byte reason);

	private ViewMemberData createViewMemberData(long sessionid, short classindex, ViewVariableData data) {
		return new ViewMemberData(sessionid,
				new ViewVariableData(
						(byte) (subscribes.get(classindex).get((byte) (data.index & 0x7f)) | (data.index & 0x80)),
						data.field, data.data, data.dataremoved));
	}

	@Override
	void assembleProtocolData(SyncViewToClients protocol, StringStream sstream) {
		mbcache.forEach(i -> protocol.members.add(createViewMemberData(i.sessionid, i.classindex, i.data.data)));
		if (sstream != null) {
			sstream.append(":P");
			mbcache.stream().filter(i -> !i.data.text.isEmpty())
					.forEach(
							i -> sstream.marshal(i.sessionid)
									.append(getTouch(
											subscribes.get(i.classindex).get((byte) (i.data.data.index & 0x7f)))
											+ i.data.text.substring(i.data.text.indexOf('?', 1))));
			protocol.stringdata = sstream.toString(":");
		}
		mbcache.clear();
	}

	void onSessionViewUpdate(long sessionid, short classindex, DataItem item) {
		hashSchedule(() -> {
			Map<Integer, DataItem> recent = members.get(sessionid);
			if (recent == null) {
				if (inprogress.contains(sessionid))
					hashSchedule(() -> onSessionViewUpdate(sessionid, classindex, item));
				return;
			}
			if (item.data.index >= 0) {
				if (item.data.data.size() > 0)
					recent.put(((int) classindex << 16) | item.data.index, item);
			} else {
				recent.remove(((int) classindex << 16) | (item.data.index & 0x7f));
			}
			if (item.additional == null)
				mbcache.add(new MemberData(sessionid, classindex, item));
			else
				item.additional.forEach(i -> mbcache.add(new MemberData(sessionid, classindex, i)));
			delaySchedule(() -> hashSchedule(() -> {
				if (mbcache.isEmpty() || isClosed())
					return;
				final SyncViewToClients protocol = protocol(SyncViewToClients.DT_TEMPORARY_DATA);
				protocol.sessionids.addAll(members.keySet());
				assembleProtocolData(protocol, prepareStringHeader(protocol));
				try {
					getViewContext().syncViewToClients(protocol);
				} catch (Throwable e) {
					if (Trace.isErrorEnabled())
						Trace.error("sendMemberData view = " + this, e);
				}
			}));
		});
	}

	private void unsubscribe(long sessionid) {
		ViewContext context = getViewContext();
		subscribes.keySet().forEach(k -> {
			SessionView sview = context.findSessionView(sessionid, k);
			if (sview != null)
				sview.unsubscribe(this);
		});
	}

	void close(final long sessionid) {
		hashSchedule(() -> {
			if (isClosed())
				return;
			inprogress.remove(sessionid);
			if (members.remove(sessionid) == null)
				return;
			flush();
			sendDetachNotify(members.keySet(), sessionid, DetachReasonClose);
			hashSchedule(sessionid, () -> unsubscribe(sessionid));
			onDetached(sessionid, DetachReasonClose);
		});
	}

	@Override
	void close() {
		hashSchedule(() -> {
			inprogress.clear();
			flush();
			Set<Long> sessionids = new HashSet<>(members.keySet());
			members.clear();
			sendCloseToClients(sessionids);
			sessionids.forEach(i -> {
				hashSchedule(i, () -> {
					ViewSession vs = getViewContext().getViewSession(i);
					if (null != vs)
						vs.remove(this);
					unsubscribe(i);
				});
				onDetached(i, DetachReasonClose);
			});
			super.close();
		});
	}

	private SyncViewToClients protocol(int synctype) {
		final SyncViewToClients protocol = new SyncViewToClients();
		protocol.providerid = getProviderId();
		protocol.classindex = getClassIndex();
		protocol.instanceindex = getInstanceIndex();
		protocol.synctype = (byte) synctype;
		return protocol;
	}

	private void sendInitViewToClient(long sessionid) {
		final SyncViewToClients protocol = protocol(SyncViewToClients.DT_TEMPORARY_INIT_DATA);
		protocol.sessionids.add(sessionid);
		final Collection<DataItem> items = getRecentDataItem();
		items.forEach(i -> protocol.vardatas.add(i.data));
		members.forEach((k, v) -> v
				.forEach((k0, v0) -> protocol.members.add(createViewMemberData(k, (short) (k0 >> 16), v0.data))));
		StringStream sstream = prepareStringHeader(protocol);
		if (sstream != null) {
			items.forEach(i -> sstream.append(i.text));
			sstream.append(":P");
			members.forEach((k, v) -> {
				if (v.isEmpty())
					sstream.marshal(k).append('U');
				else
					v.forEach((k0, v0) -> sstream.marshal(k)
							.append(getTouch(subscribes.get((short) (k0 >> 16)).get((byte) (k0 & 0x7f)))
									+ v0.text.substring(v0.text.indexOf('?', 1))));
			});
			protocol.stringdata = sstream.toString(":");
		}
		try {
			getViewContext().syncViewToClients(protocol);
		} catch (Throwable e) {
			if (Trace.isErrorEnabled())
				Trace.error("sendFullViewToClient view = " + this, e);
		}
	}

	private void sendAttachNotify(long sessionid, Map<Short, Collection<DataItem>> initData) {
		final SyncViewToClients protocol = protocol(SyncViewToClients.DT_TEMPORARY_ATTACH);
		protocol.sessionids.addAll(members.keySet());
		initData.forEach((k, v) -> v.forEach(v0 -> protocol.members.add(createViewMemberData(sessionid, k, v0.data))));
		StringStream sstream = prepareStringHeader(protocol);
		if (sstream != null) {
			sstream.append(":P");
			if (initData.values().stream().flatMap(v -> v.stream()).findAny().isPresent())
				initData.forEach((k,
						v) -> v.forEach(v0 -> sstream.marshal(sessionid)
								.append(getTouch(subscribes.get(k).get((byte) (v0.data.index & 0x7f)))
										+ v0.text.substring(v0.text.indexOf('?', 1)))));
			else
				sstream.marshal(sessionid).append('U');
			protocol.stringdata = sstream.toString(":");
		}
		try {
			getViewContext().syncViewToClients(protocol);
		} catch (Throwable e) {
			if (Trace.isErrorEnabled())
				Trace.error("sendAttachNotify view = " + this, e);
		}
	}

	private void sendCloseToClients(Collection<Long> sids) {
		if (sids.isEmpty())
			return;
		final SyncViewToClients protocol = protocol(SyncViewToClients.DT_TEMPORARY_CLOSE);
		protocol.sessionids.addAll(sids);
		StringStream ss = prepareStringHeader(protocol);
		if (ss != null)
			protocol.stringdata = ss.append(":M:").toString();
		try {
			getViewContext().syncViewToClients(protocol);
		} catch (Throwable e) {
			if (Trace.isErrorEnabled())
				Trace.error("sendCloseToClient view =" + this, e);
		}
	}

	private void sendDetachNotify(Collection<Long> dests, long sessionid, byte reason) {
		if (dests.isEmpty())
			return;
		final SyncViewToClients protocol = protocol(SyncViewToClients.DT_TEMPORARY_DETACH);
		protocol.sessionids.addAll(dests);
		ViewMemberData memberData = new ViewMemberData();
		memberData.sessionid = sessionid;
		memberData.vardata = new ViewVariableData();
		memberData.vardata.index = reason;
		protocol.members.add(memberData);
		StringStream sstream = prepareStringHeader(protocol);
		if (sstream != null)
			protocol.stringdata = sstream.append(":P").marshal(sessionid).marshal(reason).toString(":");
		try {
			getViewContext().syncViewToClients(protocol);
		} catch (Throwable e) {
			if (Trace.isErrorEnabled())
				Trace.error("sendDetachNotify view = " + this, e);
		}
	}
}
